# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import flexsolve as flx
import biosteam as bst
import thermosteam as tmo
cost = bst.decorators.cost

__all__ = ('BoilerTurbogenerator', 'Boiler')

#: TODO add reference of NREL
@cost('Work', 'Turbogenerator',
      CE=551, S=42200, kW=0, cost=9500e3, n=0.60, BM=1.8)
@cost('Flow rate', 'Hot process water softener system', 
      CE=551, cost=78e3, S=235803, n=0.6, BM=1.8)
@cost('Flow rate', 'Amine addition pkg', 
      CE=551, cost=40e3, S=235803, n=0.6, BM=1.8)
@cost('Flow rate', 'Deaerator',
      CE=551, cost=305e3, S=235803, n=0.6, BM=3.0)
@cost('Flow rate', 'Boiler',
      CE=551, cost=28550e3, kW=1371, S=238686, n=0.6, BM=1.8)
@cost('Ash disposal', 'Baghouse bags',
      CE=551, cost=466183. / 4363., n=1.0, lifetime=5)
class BoilerTurbogenerator(bst.Facility):
    """
    Create a BoilerTurbogenerator object that will calculate electricity
    generation from burning the feed. It also takes into account how much
    steam is being produced, and the required cooling utility of the turbo
    generator. All capital cost correlations are based on [1]_.
    
    Parameters
    ----------
    ins : 
        * [0] Liquid/solid feed that will be burned.
        * [1] Gas feed that will be burned.
        * [2] Make-up water. 
        * [3] Natural gas/fuel to satisfy steam and electricity demand.
        * [4] Lime for flue gas desulfurization.
        * [5] Boiler chemicals.
        * [6] Air or oxygen-rich gas.
    outs : 
        * [0] Total emissions produced.
        * [1] Blowdown water.
        * [2] Ash disposal.
    boiler_efficiency : float, optional
        Fraction of heat transferred to steam. Defaults to 0.8.
    turbo_generator_efficiency : float, optional
        Fraction of steam heat converted to electricity. Defaults to 0.85.
    agent : UtilityAgent, optional
        Steam produced. Defaults to low pressure steam.
    other_agents = () : Iterable[UtilityAgent], optional
        Other steams produced. Defaults to all other heating agents.
    fuel_source : str, optional
        Name fuel used to satisfy steam and electricity demand. Defaults to 'CH4'.
    fuel_price : float, optional
        Price of natural gas [USD/kg]. Same as `bst.stream_utility_prices['Natural gas']`,
        which defaults to 0.218.
    ash_disposal_price : float, optional
        Price of disposing ash [USD/kg]. Same as `bst.stream_utility_prices['Ash disposal']`,
        which defaults to -0.0318.
    satisfy_system_electricity_demand : bool, optional
        Whether to purchase natural gas to satisfy system electricity demand
        if there is not enough heat from process feeds (i.e., inlets 0 and 1).
        If True, natural gas is purchased to satisfy system heat and electricity demand
        when there is not enough heat from the feed and gas.
        If False, natural gas is only purchased to satisfy system heat demand
        and electricity will be purchased from the grid if there is not
        enough heat from the feeds.
        In either case, if there is excess heat from the process feeds,
        electricity will still be produced.
    boiler_efficiency_basis : str, optional
        Basis of boiler efficiency. Defaults to 'LHV' (i.e., lower heating value).
        'HHV' (i.e., higher heating value) is also a valid basis. 
        
    Examples
    --------
    Create a boiler-turbogenerator system that uses sugarcane bagasse to 
    produce steam for a distillation unit and any excess steam for surplus electricity:
    
    >>> import biosteam as bst
    >>> from biorefineries import cane
    >>> chemicals = cane.create_sugarcane_chemicals()
    >>> chemicals.define_group(
    ...     name='Fiber',
    ...     IDs=['Cellulose', 'Hemicellulose', 'Lignin'],
    ...     composition=[0.4704 , 0.2775, 0.2520],
    ...     wt=True, # Composition is given as weight
    ... )
    >>> bst.settings.set_thermo(chemicals)
    >>> dilute_ethanol = bst.Stream('dilute_ethanol', Water=1390, Ethanol=590)
    >>> bagasse = bst.Stream('bagasse', Water=0.4, Fiber=0.6, total_flow=8e4, units='kg/hr')
    >>> with bst.System('sys') as sys:
    ...     D1 = bst.BinaryDistillation('D1', ins=dilute_ethanol, Lr=0.999, Hr=0.89, k=1.25, LHK=('Ethanol', 'Water'))
    ...     BT = bst.BoilerTurbogenerator('BT')
    ...     BT.ins[0] = bagasse
    >>> sys.simulate()
    >>> BT.show()
    BoilerTurbogenerator: BT
    ins...
    [0] bagasse  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water          1.78e+03
                        Cellulose      139
                        Hemicellulose  101
                        Lignin         79.5
    [1] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow: 0
    [2] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  488
    [3] -  
        phase: 'g', T: 288.71 K, P: 101560 Pa
        flow: 0
    [4] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow: 0
    [5] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Ash  0.567
    [6] -  
        phase: 'g', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): O2  9.85e+03
                        N2  4.23e+04
    outs...
    [0] emissions  
        phase: 'g', T: 394.15 K, P: 101325 Pa
        flow (kmol/hr): Water  3.19e+03
                        CO2    1.98e+03
                        O2     7.84e+03
                        N2     4.23e+04
    [1] rejected_water_and_blowdown  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  488
    [2] ash_disposal  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  0.00944
                        Ash    0.567
    
    >>> BT.results() # Steam and electricity are produced, so costs are negative
    Boiler turbogenerator                                      Units        BT
    Electricity           Power                                   kW -1.31e+05
                          Cost                                USD/hr -1.02e+04
    Low pressure steam    Duty                                 kJ/hr -7.32e+07
                          Flow                               kmol/hr -1.89e+03
                          Cost                                USD/hr      -450
    Cooling water         Duty                                 kJ/hr -8.42e+07
                          Flow                               kmol/hr  5.75e+04
                          Cost                                USD/hr      28.1
    Fuel (inlet)          Flow                                 kg/hr         0
                          Cost                                USD/hr         0
    Ash disposal (outlet) Flow                                 kg/hr     0.737
                          Cost                                USD/hr    0.0234
    Design                Work                                    kW  1.33e+05
                          Flow rate                            kg/hr  2.93e+05
                          Ash disposal                         kg/hr     0.737
    Purchase cost         Baghouse bags                          USD      81.1
                          Boiler                                 USD  3.33e+07
                          Deaerator                              USD  3.58e+05
                          Amine addition pkg                     USD  4.69e+04
                          Hot process water softener system      USD  9.16e+04
                          Turbogenerator                         USD  1.94e+07
    Total purchase cost                                          USD  5.32e+07
    Utility cost                                              USD/hr -1.07e+04
    
    Notes
    -----
    The flow rate of natural gas, lime, and boiler chemicals (streams 3-5)
    is set by the BoilerTurbogenerator object during simulation.
    
    References
    ----------
    .. [1] Humbird, D., Davis, R., Tao, L., Kinchin, C., Hsu, D., Aden, A.,
        Dudgeon, D. (2011). Process Design and Economics for Biochemical 
        Conversion of Lignocellulosic Biomass to Ethanol: Dilute-Acid 
        Pretreatment and Enzymatic Hydrolysis of Corn Stover
        (No. NREL/TP-5100-47764, 1013269). https://doi.org/10.2172/1013269
    
    """
    ticket_name = 'BT'
    network_priority = 1
    boiler_blowdown = 0.03
    # Reverse osmosis (RO) typically rejects 25% of water, but the boiler-feed water is assumed to come after RO.
    # Setting this parameter to a fraction more than zero effectively assumes that this unit accounts for reverse osmosis.
    RO_rejection = 0 
    _N_ins = 7
    _N_outs = 3
    _units = {'Flow rate': 'kg/hr',
              'Work': 'kW',
              'Ash disposal': 'kg/hr'}
    
    def __init__(self, ID='', ins=None, 
                 outs=('emissions',
                       'rejected_water_and_blowdown',
                       'ash_disposal'),
                 thermo=None, *,
                 boiler_efficiency=None,
                 turbogenerator_efficiency=None,
                 side_steam=None,
                 agent=None,
                 other_agents=None,
                 fuel_price=None,
                 natural_gas_price=None,
                 ash_disposal_price=None,
                 T_emissions=None,
                 CO2_emissions_concentration=None,
                 satisfy_system_electricity_demand=None,
                 boiler_efficiency_basis=None,
                 fuel_source=None,
                 oxygen_rich_gas_composition=None,
        ):
        if boiler_efficiency_basis is None: boiler_efficiency_basis = 'LHV'
        if boiler_efficiency is None: boiler_efficiency = 0.80
        if turbogenerator_efficiency is None: turbogenerator_efficiency = 0.85
        if satisfy_system_electricity_demand is None: satisfy_system_electricity_demand = True
        if fuel_source is None: fuel_source = 'CH4'
        if oxygen_rich_gas_composition is None: oxygen_rich_gas_composition = dict(O2=21, N2=79, phase='g', units='kg/hr')
        if CO2_emissions_concentration is None: CO2_emissions_concentration = 0.055  # Usually between 4 - 7 for biomass and natural gas (https://www.sciencedirect.com/science/article/pii/S0957582021005127)
        bst.Facility.__init__(self, ID, ins, outs, thermo)
        settings = bst.settings
        self.boiler_efficiency_basis = boiler_efficiency_basis
        self.agent = agent = agent or settings.get_heating_agent('low_pressure_steam')
        self.fuel_source = fuel_source
        self.define_utility('Fuel', self.fuel)
        self.define_utility('Ash disposal', self.ash_disposal)
        self.boiler_efficiency = boiler_efficiency
        self.turbogenerator_efficiency = turbogenerator_efficiency
        self.steam_utilities = []
        self.steam_demand = agent.to_stream()
        self.side_steam = side_steam
        self.other_agents = [i for i in settings.heating_agents if i is not agent] if other_agents is None else other_agents
        self.CO2_emissions_concentration = CO2_emissions_concentration
        self.oxygen_rich_gas_composition = oxygen_rich_gas_composition 
        if T_emissions is not None: 
            raise ValueError('setting T_emissions is not yet implemented')
            # T_emissions should dictate the efficiency of either the boiler or the turbogenerator.
            # Note that T_emissions is tied to the energy balance.
            self.T_emissions = T_emissions
        if natural_gas_price is not None: 
            self.fuel_price = natural_gas_price
        elif fuel_price is not None: 
            self.fuel_price = fuel_price
        if ash_disposal_price is not None: self.ash_disposal_price = ash_disposal_price
        self.satisfy_system_electricity_demand = satisfy_system_electricity_demand
      
    def _get_desulfurization_rxn_and_coreactant(self):
        try:
            return self.desulfurization_reaction, self._ID_lime
        except:
            chemicals = self.chemicals
            CAS_lime = '1305-62-0'
            has_common_name = 'Ca(OH)2' in chemicals
            if CAS_lime in chemicals or has_common_name:
                if not has_common_name: chemicals.set_synonym(CAS_lime, 'Ca(OH)2')
                self.desulfurization_reaction = rxn = tmo.Reaction(
                    'SO2 + Ca(OH)2 + 0.5 O2 -> CaSO4 + H2O', 'SO2', 0.92, chemicals
                )
                self._ID_lime = ID = 'Ca(OH)2'
                return rxn, ID
            CAS_lime = '1305-78-8'
            has_common_name = 'CaO' in chemicals
            if CAS_lime in chemicals or has_common_name:
                if not has_common_name: chemicals.set_synonym(CAS_lime, 'CaO')
                self.desulfurization_reaction = rxn = tmo.Reaction(
                    'SO2 + CaO + 0.5 O2 -> CaSO4', 'SO2', 0.92, chemicals
                )
                self._ID_lime = ID = 'CaO'
                return rxn, ID
        raise RuntimeError(
            "lime is required for boiler, but no chemical 'CaO' or 'Ca(OH)2' "
            "available in thermodynamic property package"
        )
        
    @property
    def emissions(self):
        return self.outs[0]
    
    @property
    def blowdown_water(self):
        return self.outs[1]
    
    @property
    def makeup_water(self):
        """[Stream] Makeup water due to boiler blowdown."""
        return self.ins[2]
    
    @property
    def fuel(self):
        """[Stream] Fuel used to satisfy steam and electricity requirements."""
        return self.ins[3]
    natural_gas = fuel
    
    @property
    def air(self):
        """[Stream] Air or oxygen rich gas used to supply oxygen for combustion."""
        return self.ins[6]
    oxygen_rich_gas = air
    
    @property
    def ash_disposal(self):
        """[Stream] Ash disposal."""
        return self.outs[2]
    
    @property
    def fuel_price(self):
        """[Float] Price of natural gas, same as `bst.stream_utility_prices['Natural gas']`."""
        return bst.stream_utility_prices['Fuel']
    @fuel_price.setter
    def fuel_price(self, new_price):
        bst.stream_utility_prices['Fuel'] = new_price
    natural_gas_price = fuel_price
    
    @property
    def ash_disposal_price(self):
        """[Float] Price of ash disposal, same as `bst.stream_utility_prices['Ash disposal']`."""
        return bst.stream_utility_prices['Ash disposal']
    
    @ash_disposal_price.setter
    def ash_disposal_price(self, ash_disposal_price):
        bst.stream_utility_prices['Ash disposal'] = ash_disposal_price
    
    def _run(self): pass

    def _load_utility_agents(self):
        steam_utilities = self.steam_utilities
        steam_utilities.clear()
        agent = self.agent
        units = self.other_units
        for agent in (*self.other_agents, agent):
            ID = agent.ID
            for u in units:
                for hu in u.heat_utilities:
                    agent = hu.agent
                    if agent and agent.ID == ID:
                        steam_utilities.append(hu)
        self.electricity_demand = max(sum([u.power_utility.rate for u in units]), 0)
    
    def _design(self):
        B_eff = self.boiler_efficiency
        TG_eff = self.turbogenerator_efficiency
        steam_demand = self.steam_demand
        Design = self.design_results
        chemicals = self.chemicals
        self._load_utility_agents()
        mol_steam = sum([i.flow for i in self.steam_utilities])
        feed_solids, feed_gas, makeup_water, fuel, lime, chems, oxygen_rich_gas = self.ins
        oxygen_rich_gas.empty()
        if self.fuel_source == 'CH4':
            fuel.phase = 'g'
            fuel.set_property('T', 60, 'degF')
            fuel.set_property('P', 14.73, 'psi')
        emissions, blowdown_water, ash_disposal = self.outs
        if not lime.price:
            lime.price = 0.19937504680689402
        if not chems.price:
            chems.price = 4.995862254032183
        H_steam =  sum([i.duty for i in self.steam_utilities])
        side_steam = self.side_steam
        if side_steam: 
            H_steam += side_steam.H
            mol_steam += side_steam.F_mol
        steam_demand.imol['7732-18-5'] = mol_steam 
        self.combustion_reactions = combustion_rxns = chemicals.get_combustion_reactions()
        non_empty_feeds = [i for i in (feed_solids, feed_gas) if not i.isempty()]
        boiler_efficiency_basis = self.boiler_efficiency_basis
        fuel_source = self.fuel_source
        def calculate_excess_electricity_at_natual_gas_flow(fuel_flow):
            if fuel_flow:
                fuel_flow = abs(fuel_flow)
                fuel.imol[fuel_source] = fuel_flow
            else:
                fuel.empty()
            if boiler_efficiency_basis == 'LHV':
                H_combustion = fuel.LHV
                for feed in non_empty_feeds: H_combustion += feed.LHV
            elif boiler_efficiency_basis == 'HHV':
                H_combustion = fuel.HHV
                for feed in non_empty_feeds: H_combustion += feed.HHV
            else:
                raise ValueError(
                    f"invalid boiler efficiency basis {boiler_efficiency_basis}; "
                    f"valid values include 'LHV', or 'HHV'"
                )
            self.H_content = H_content = B_eff * H_combustion 
            self.H_loss_to_emissions = H_combustion - H_content
            H_electricity = H_content - H_steam # Heat available for the turbogenerator
            electricity = H_electricity * TG_eff  # Electricity produced
            self.cooling_duty = electricity - H_electricity
            Design['Work'] = work = electricity / 3600
            duty_over_mol = 39000 # kJ / mol-superheated steam 
            self.total_steam = H_content / duty_over_mol #: [float] Total steam produced by the boiler (kmol/hr)
            Design['Flow rate'] = flow_rate = self.total_steam * 18.01528
            
            if self.satisfy_system_electricity_demand:
                boiler = self.cost_items['Boiler']
                rate_boiler = boiler.kW * flow_rate / boiler.S
                return work - self.electricity_demand - rate_boiler
            else:
                return work
        
        self._excess_electricity_without_fuel = excess_electricity = calculate_excess_electricity_at_natual_gas_flow(0)
        if excess_electricity < 0:
            f = calculate_excess_electricity_at_natual_gas_flow
            lb = 0.
            fuel.imol[fuel_source] = 1
            ub = - excess_electricity * 3600 / fuel.LHV
            while f(ub) < 0.: 
                lb = ub
                ub *= 2
            flx.IQ_interpolation(f, lb, ub, xtol=1, ytol=1)
        
        if self.cooling_duty > 0.: 
            # In the event that no electricity is produced and the solver
            # solution for natural gas is slightly below the requirement for steam
            # (this would lead to a positive duty).
            self.cooling_duty = 0.
            Design['Work'] = 0.
        
        emissions.T = 298.15 # Will be updated later with the energy balance
        emissions.P = 101325
        emissions.phase = 'g'
        emissions_mol = emissions.mol
        emissions_mol[:] = fuel.mol
        for feed in non_empty_feeds: emissions_mol[:] += feed.mol
        combustion_rxns.force_reaction(emissions_mol)
        O2_consumption = -emissions.imol['O2']
        oxygen_rich_gas.reset_flow(**self.oxygen_rich_gas_composition)
        z_O2 = oxygen_rich_gas.imol['O2'] / oxygen_rich_gas.F_mol
        oxygen_rich_gas.F_mol = O2_consumption / z_O2
        emissions_mol += oxygen_rich_gas.mol
        F_emissions = emissions.F_mass
        z_CO2 = emissions.imass['CO2'] / F_emissions
        z_CO2_target = self.CO2_emissions_concentration
        if z_CO2 > z_CO2_target:
            F_emissions_new = z_CO2 * F_emissions / z_CO2_target
            dF_emissions = F_emissions_new - F_emissions
            oxygen_rich_gas.F_mass = F_mass_O2_new = oxygen_rich_gas.F_mass + dF_emissions
            emissions_mol += oxygen_rich_gas.mol * (dF_emissions / F_mass_O2_new)
        emissions.H += self.H_loss_to_emissions
        hu_cooling = bst.HeatUtility()
        hu_cooling(self.cooling_duty, steam_demand.T)
        hus_heating = bst.HeatUtility.sum_by_agent(self.steam_utilities)
        for hu in hus_heating: hu.reverse()
        self.heat_utilities = [*hus_heating, hu_cooling]
        water_index = chemicals.index('7732-18-5')
        makeup_water.mol[water_index] = blowdown_water.mol[water_index] = (
                self.total_steam * self.boiler_blowdown * 1 / (1 - self.RO_rejection)   
        )
        ash_IDs = [i.ID for i in self.chemicals if not i.formula]
        emissions_mol = emissions.mol
        SO2_produced = 0
        if 'SO2' in chemicals:
            SO2_produced += emissions.imol['SO2']
        if 'CaSO4' in chemicals:
            SO2_produced += emissions.imol['CaSO4']
            ash_IDs.append('CaSO4')
        if SO2_produced: 
            rxn, ID_lime = self._get_desulfurization_rxn_and_coreactant()
            # FGD lime scaled based on SO2 generated,	
            # 20% stoichiometric excess based on P52 of ref [1]
            rxn.force_reaction(emissions)
            lime.imol[ID_lime] = lime_mol = SO2_produced * 1.2
            emissions_mol.remove_negatives()
        else:
            lime.empty()
        # About 0.4536 kg/hr of boiler chemicals are needed per 234484 kg/hr steam produced
        chems.imol['Ash'] = boiler_chems = 1.9345e-06 * Design['Flow rate']
        ash_disposal.empty()
        ash_disposal.copy_flow(emissions, IDs=tuple(ash_IDs), remove=True)
        ash_disposal.imol['Ash'] += boiler_chems
        dry_ash = ash_disposal.F_mass
        moisture = min(emissions.imass['Water'], dry_ash * 0.3) # ~20% moisture
        ash_disposal.imass['Water'] = moisture
        emissions.imass['Water'] -= moisture
        Design['Ash disposal'] = dry_ash + moisture
        if SO2_produced:
            if ID_lime == 'Ca(OH)2': # Ca(OH)2
                lime.imol['Water'] = 4 * lime_mol # Its a slurry
            else: # CaO
                lime.imol['Water'] = 5 * lime_mol 
        
    def _cost(self):
        self._decorated_cost()
        self.power_utility.production = self.design_results['Work']

# Simulation of ethanol production from sugarcane
# in Brazil: economic study of an autonomous
# distillery
# Marina O.S. Diasa,b, Marcelo P. Cunhaa, Charles D.F. Jesusa, Mirna I.G.
# Scandiffioa, Carlos E.V. Rossella,b, Rubens Maciel Filhob, Antonio Bonomia
# a CTBE – Bioethanol Science and Technology National Laboratory, PO Box 6170 –
# CEP 13083-970, Campinas – SP, Brazil, marina.dias@bioetanol.org.br
# b School of Chemical Engineering, University of Campinas, PO Box 6066 – CEP
# 13083-970, Campinas – SP, Brazil
# LHV = 7565 # Bagasse lower heating value (kJ/kg)

@cost('Flow rate', 'Hot process water softener system', 
      CE=551, cost=78e3, S=235803, n=0.6, BM=1.8)
@cost('Flow rate', 'Amine addition pkg', 
      CE=551, cost=40e3, S=235803, n=0.6, BM=1.8)
@cost('Flow rate', 'Deaerator',
      CE=551, cost=305e3, S=235803, n=0.6, BM=3.0)
@cost('Flow rate', 'Boiler',
      CE=551, cost=28550e3, kW=1371, S=238686, n=0.6, BM=1.8)
@cost('Ash disposal', 'Baghouse bags',
      CE=551, cost=466183. / 4363., n=1.0, lifetime=5)
class Boiler(bst.Facility):
    """
    Create a Boiler object that will calculate steam
    generation from burning the feed.
    All capital cost correlations are based on [1]_.
    
    Parameters
    ----------
    ins : 
        * [0] Liquid/solid feed that will be burned.
        * [1] Gas feed that will be burned.
        * [2] Make-up water. 
        * [3] Natural gas/fuel to satisfy steam and electricity demand.
        * [4] Lime for flue gas desulfurization.
        * [5] Boiler chemicals.
        * [6] Air or oxygen-rich gas.
    outs : 
        * [0] Total emissions produced.
        * [1] Blowdown water.
        * [2] Ash disposal.
    boiler_efficiency : float, optional
        Fraction of heat transferred to steam. Defaults to 0.8.
    agent : UtilityAgent, optional
        Steam produced. Defaults to low pressure steam.
    other_agents = () : Iterable[UtilityAgent], optional
        Other steams produced. Defaults to all other heating agents.
    fuel_source : str, optional
        Name fuel used to satisfy steam and electricity demand. Defaults to 'CH4'.
    fuel_price : float, optional
        Price of natural gas [USD/kg]. Same as `bst.stream_utility_prices['Natural gas']`,
        which defaults to 0.218.
    ash_disposal_price : float, optional
        Price of disposing ash [USD/kg]. Same as `bst.stream_utility_prices['Ash disposal']`,
        which defaults to -0.0318.
    satisfy_system_steam_demand : bool, optional
        Whether to purchase natural gas to satisfy system steam demand
        if there is not enough heat from process feeds (i.e., inlets 0 and 1).
        If True, natural gas is purchased to satisfy system heat demand
        when there is not enough heat from the process feeds.
        In either case, if there is excess heat from the process feeds,
        excess steam will be produced.
    boiler_efficiency_basis : str, optional
        Basis of boiler efficiency. Defaults to 'LHV' (i.e., lower heating value).
        'HHV' (i.e., higher heating value) is also a valid basis. 
        
    Examples
    --------
    Create a boiler-turbogenerator system that uses sugarcane bagasse to 
    produce steam for a distillation unit and any excess steam for surplus electricity:
    
    >>> import biosteam as bst
    >>> from biorefineries import cane
    >>> chemicals = cane.create_sugarcane_chemicals()
    >>> chemicals.define_group(
    ...     name='Fiber',
    ...     IDs=['Cellulose', 'Hemicellulose', 'Lignin'],
    ...     composition=[0.4704 , 0.2775, 0.2520],
    ...     wt=True, # Composition is given as weight
    ... )
    >>> bst.settings.set_thermo(chemicals)
    >>> dilute_ethanol = bst.Stream('dilute_ethanol', Water=1390, Ethanol=590)
    >>> bagasse = bst.Stream('bagasse', Water=0.4, Fiber=0.6, total_flow=8e4, units='kg/hr')
    >>> with bst.System('sys') as sys:
    ...     D1 = bst.BinaryDistillation('D1', ins=dilute_ethanol, Lr=0.999, Hr=0.89, k=1.25, LHK=('Ethanol', 'Water'))
    ...     B = bst.Boiler('B')
    ...     B.ins[0] = bagasse
    >>> sys.simulate()
    >>> B.show()
    Boiler: B
    ins...
    [0] bagasse  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water          1.78e+03
                        Cellulose      139
                        Hemicellulose  101
                        Lignin         79.5
    [1] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow: 0
    [2] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  488
    [3] -  
        phase: 'g', T: 288.71 K, P: 101560 Pa
        flow: 0
    [4] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow: 0
    [5] -  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Ash  0.567
    [6] -  
        phase: 'g', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): O2  9.85e+03
                        N2  4.23e+04
    outs...
    [0] emissions  
        phase: 'g', T: 394.15 K, P: 101325 Pa
        flow (kmol/hr): Water  3.19e+03
                        CO2    1.98e+03
                        O2     7.84e+03
                        N2     4.23e+04
    [1] rejected_water_and_blowdown  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  488
    [2] ash_disposal  
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  0.00944
                        Ash    0.567
    
    >>> B.results() # Steam is produced, so utility cost is negative
    Boiler                                                     Units         B
    Electricity           Power                                   kW  1.68e+03
                          Cost                                USD/hr       132
    Low pressure steam    Duty                                 kJ/hr -7.32e+07
                          Flow                               kmol/hr -1.89e+03
                          Cost                                USD/hr      -450
    Fuel (inlet)          Flow                                 kg/hr         0
                          Cost                                USD/hr         0
    Ash disposal (outlet) Flow                                 kg/hr     0.737
                          Cost                                USD/hr    0.0234
    Design                Flow rate                            kg/hr  2.93e+05
                          Ash disposal                         kg/hr     0.737
    Purchase cost         Baghouse bags                          USD      81.1
                          Boiler                                 USD  3.33e+07
                          Deaerator                              USD  3.58e+05
                          Amine addition pkg                     USD  4.69e+04
                          Hot process water softener system      USD  9.16e+04
    Total purchase cost                                          USD  3.38e+07
    Utility cost                                              USD/hr      -319
    
    Notes
    -----
    The flow rate of natural gas, lime, and boiler chemicals (streams 3-5)
    is set by the BoilerTurbogenerator object during simulation.
    
    References
    ----------
    .. [1] Humbird, D., Davis, R., Tao, L., Kinchin, C., Hsu, D., Aden, A.,
        Dudgeon, D. (2011). Process Design and Economics for Biochemical 
        Conversion of Lignocellulosic Biomass to Ethanol: Dilute-Acid 
        Pretreatment and Enzymatic Hydrolysis of Corn Stover
        (No. NREL/TP-5100-47764, 1013269). https://doi.org/10.2172/1013269
    
    """
    ticket_name = 'B'
    network_priority = 1
    boiler_blowdown = 0.03
    # Reverse osmosis (RO) typically rejects 25% of water, but the boiler-feed water is assumed to come after RO.
    # Setting this parameter to a fraction more than zero effectively assumes that this unit accounts for reverse osmosis.
    RO_rejection = 0 
    _N_ins = 7
    _N_outs = 3
    _units = {'Flow rate': 'kg/hr',
              'Work': 'kW',
              'Ash disposal': 'kg/hr'}
    
    def __init__(self, ID='', ins=None, 
                 outs=('emissions',
                       'rejected_water_and_blowdown',
                       'ash_disposal'),
                 thermo=None, *,
                 boiler_efficiency=None,
                 side_steam=None,
                 agent=None,
                 other_agents=None,
                 fuel_price=None,
                 natural_gas_price=None,
                 ash_disposal_price=None,
                 T_emissions=None,
                 CO2_emissions_concentration=None,
                 satisfy_system_steam_demand=None,
                 boiler_efficiency_basis=None,
                 fuel_source=None,
                 oxygen_rich_gas_composition=None,
        ):
        if boiler_efficiency_basis is None: boiler_efficiency_basis = 'LHV'
        if boiler_efficiency is None: boiler_efficiency = 0.80
        if satisfy_system_steam_demand is None: satisfy_system_electricity_demand = True
        if fuel_source is None: fuel_source = 'CH4'
        if oxygen_rich_gas_composition is None: oxygen_rich_gas_composition = dict(O2=21, N2=79, phase='g', units='kg/hr')
        if CO2_emissions_concentration is None: CO2_emissions_concentration = 0.055  # Usually between 4 - 7 for biomass and natural gas (https://www.sciencedirect.com/science/article/pii/S0957582021005127)
        bst.Facility.__init__(self, ID, ins, outs, thermo)
        settings = bst.settings
        self.boiler_efficiency_basis = boiler_efficiency_basis
        self.agent = agent = agent or settings.get_heating_agent('low_pressure_steam')
        self.fuel_source = fuel_source
        self.define_utility('Fuel', self.fuel)
        self.define_utility('Ash disposal', self.ash_disposal)
        self.boiler_efficiency = boiler_efficiency
        self.steam_utilities = []
        self.steam_demand = agent.to_stream()
        self.side_steam = side_steam
        self.other_agents = [i for i in settings.heating_agents if i is not agent] if other_agents is None else other_agents
        self.CO2_emissions_concentration = CO2_emissions_concentration
        self.oxygen_rich_gas_composition = oxygen_rich_gas_composition 
        if T_emissions is not None: 
            raise ValueError('setting T_emissions is not yet implemented')
            # T_emissions should dictate the efficiency of either the boiler or the turbogenerator.
            # Note that T_emissions is tied to the energy balance.
            self.T_emissions = T_emissions
        if natural_gas_price is not None: 
            self.fuel_price = natural_gas_price
        elif fuel_price is not None: 
            self.fuel_price = fuel_price
        if ash_disposal_price is not None: self.ash_disposal_price = ash_disposal_price
        self.satisfy_system_steam_demand = satisfy_system_steam_demand
      
    def _get_desulfurization_rxn_and_coreactant(self):
        try:
            return self.desulfurization_reaction, self._ID_lime
        except:
            chemicals = self.chemicals
            CAS_lime = '1305-62-0'
            has_common_name = 'Ca(OH)2' in chemicals
            if CAS_lime in chemicals or has_common_name:
                if not has_common_name: chemicals.set_synonym(CAS_lime, 'Ca(OH)2')
                self.desulfurization_reaction = rxn = tmo.Reaction(
                    'SO2 + Ca(OH)2 + 0.5 O2 -> CaSO4 + H2O', 'SO2', 0.92, chemicals
                )
                self._ID_lime = ID = 'Ca(OH)2'
                return rxn, ID
            CAS_lime = '1305-78-8'
            has_common_name = 'CaO' in chemicals
            if CAS_lime in chemicals or has_common_name:
                if not has_common_name: chemicals.set_synonym(CAS_lime, 'CaO')
                self.desulfurization_reaction = rxn = tmo.Reaction(
                    'SO2 + CaO + 0.5 O2 -> CaSO4', 'SO2', 0.92, chemicals
                )
                self._ID_lime = ID = 'CaO'
                return rxn, ID
        raise RuntimeError(
            "lime is required for boiler, but no chemical 'CaO' or 'Ca(OH)2' "
            "available in thermodynamic property package"
        )
        
    @property
    def emissions(self):
        return self.outs[0]
    
    @property
    def blowdown_water(self):
        return self.outs[1]
    
    @property
    def makeup_water(self):
        """[Stream] Makeup water due to boiler blowdown."""
        return self.ins[2]
    
    @property
    def fuel(self):
        """[Stream] Fuel used to satisfy steam and electricity requirements."""
        return self.ins[3]
    natural_gas = fuel
    
    @property
    def air(self):
        """[Stream] Air or oxygen rich gas used to supply oxygen for combustion."""
        return self.ins[6]
    oxygen_rich_gas = air
    
    @property
    def ash_disposal(self):
        """[Stream] Ash disposal."""
        return self.outs[2]
    
    @property
    def fuel_price(self):
        """[Float] Price of natural gas, same as `bst.stream_utility_prices['Natural gas']`."""
        return bst.stream_utility_prices['Fuel']
    @fuel_price.setter
    def fuel_price(self, new_price):
        bst.stream_utility_prices['Fuel'] = new_price
    natural_gas_price = fuel_price
    
    @property
    def ash_disposal_price(self):
        """[Float] Price of ash disposal, same as `bst.stream_utility_prices['Ash disposal']`."""
        return bst.stream_utility_prices['Ash disposal']
    
    @ash_disposal_price.setter
    def ash_disposal_price(self, ash_disposal_price):
        bst.stream_utility_prices['Ash disposal'] = ash_disposal_price
    
    def _run(self): pass

    def _load_utility_agents(self):
        steam_utilities = self.steam_utilities
        steam_utilities.clear()
        agent = self.agent
        units = self.other_units
        for agent in (*self.other_agents, agent):
            ID = agent.ID
            for u in units:
                for hu in u.heat_utilities:
                    agent = hu.agent
                    if agent and agent.ID == ID:
                        steam_utilities.append(hu)
    
    def _design(self):
        B_eff = self.boiler_efficiency
        steam_demand = self.steam_demand
        Design = self.design_results
        chemicals = self.chemicals
        self._load_utility_agents()
        mol_steam = sum([i.flow for i in self.steam_utilities])
        feed_solids, feed_gas, makeup_water, fuel, lime, chems, oxygen_rich_gas = self.ins
        oxygen_rich_gas.empty()
        if self.fuel_source == 'CH4':
            fuel.phase = 'g'
            fuel.set_property('T', 60, 'degF')
            fuel.set_property('P', 14.73, 'psi')
        emissions, blowdown_water, ash_disposal = self.outs
        if not lime.price:
            lime.price = 0.19937504680689402
        if not chems.price:
            chems.price = 4.995862254032183
        H_steam =  sum([i.duty for i in self.steam_utilities])
        side_steam = self.side_steam
        if side_steam: 
            H_steam += side_steam.H
            mol_steam += side_steam.F_mol
        steam_demand.imol['7732-18-5'] = mol_steam 
        self.combustion_reactions = combustion_rxns = chemicals.get_combustion_reactions()
        non_empty_feeds = [i for i in (feed_solids, feed_gas) if not i.isempty()]
        boiler_efficiency_basis = self.boiler_efficiency_basis
        fuel_source = self.fuel_source
        def calculate_excess_steam_at_natual_gas_flow(fuel_flow):
            if fuel_flow:
                fuel_flow = abs(fuel_flow)
                fuel.imol[fuel_source] = fuel_flow
            else:
                fuel.empty()
            if boiler_efficiency_basis == 'LHV':
                H_combustion = fuel.LHV
                for feed in non_empty_feeds: H_combustion += feed.LHV
            elif boiler_efficiency_basis == 'HHV':
                H_combustion = fuel.HHV
                for feed in non_empty_feeds: H_combustion += feed.HHV
            else:
                raise ValueError(
                    f"invalid boiler efficiency basis {boiler_efficiency_basis}; "
                    f"valid values include 'LHV', or 'HHV'"
                )
            self.H_content = H_content = B_eff * H_combustion 
            self.H_loss_to_emissions = H_combustion - H_content
            H_excess = H_content - H_steam # Heat available for the turbogenerator
            duty_over_mol = 39000 # kJ / mol-superheated steam 
            self.total_steam = H_content / duty_over_mol #: [float] Total steam produced by the boiler (kmol/hr)
            Design['Flow rate'] = flow_rate = self.total_steam * 18.01528
            return H_excess
        
        H_excess = calculate_excess_steam_at_natual_gas_flow(0)
        if H_excess < 0:
            f = calculate_excess_steam_at_natual_gas_flow
            lb = 0.
            fuel.imol[fuel_source] = 1
            ub = - H_excess / fuel.LHV
            while f(ub) < 0.: 
                lb = ub
                ub *= 2
            flx.IQ_interpolation(f, lb, ub, xtol=1, ytol=1)
        else:
            pass
            # raise NotImplementedError('BioSTEAM Boiler cannot yet produce excess steam')
        emissions.T = 298.15 # Will be updated later with the energy balance
        emissions.P = 101325
        emissions.phase = 'g'
        emissions_mol = emissions.mol
        emissions_mol[:] = fuel.mol
        for feed in non_empty_feeds: emissions_mol[:] += feed.mol
        combustion_rxns.force_reaction(emissions_mol)
        O2_consumption = -emissions.imol['O2']
        oxygen_rich_gas.reset_flow(**self.oxygen_rich_gas_composition)
        z_O2 = oxygen_rich_gas.imol['O2'] / oxygen_rich_gas.F_mol
        oxygen_rich_gas.F_mol = O2_consumption / z_O2
        emissions_mol += oxygen_rich_gas.mol
        F_emissions = emissions.F_mass
        z_CO2 = emissions.imass['CO2'] / F_emissions
        z_CO2_target = self.CO2_emissions_concentration
        if z_CO2 > z_CO2_target:
            F_emissions_new = z_CO2 * F_emissions / z_CO2_target
            dF_emissions = F_emissions_new - F_emissions
            oxygen_rich_gas.F_mass = F_mass_O2_new = oxygen_rich_gas.F_mass + dF_emissions
            emissions_mol += oxygen_rich_gas.mol * (dF_emissions / F_mass_O2_new)
        emissions.H += self.H_loss_to_emissions
        hus_heating = bst.HeatUtility.sum_by_agent(self.steam_utilities)
        for hu in hus_heating: hu.reverse()
        self.heat_utilities = hus_heating
        water_index = chemicals.index('7732-18-5')
        makeup_water.mol[water_index] = blowdown_water.mol[water_index] = (
                self.total_steam * self.boiler_blowdown * 1 / (1 - self.RO_rejection)   
        )
        ash_IDs = [i.ID for i in self.chemicals if not i.formula]
        emissions_mol = emissions.mol
        SO2_produced = 0
        if 'SO2' in chemicals:
            SO2_produced += emissions.imol['SO2']
        if 'CaSO4' in chemicals:
            SO2_produced += emissions.imol['CaSO4']
            ash_IDs.append('CaSO4')
        if SO2_produced: 
            rxn, ID_lime = self._get_desulfurization_rxn_and_coreactant()
            # FGD lime scaled based on SO2 generated,	
            # 20% stoichiometric excess based on P52 of ref [1]
            rxn.force_reaction(emissions)
            lime.imol[ID_lime] = lime_mol = SO2_produced * 1.2
            emissions_mol.remove_negatives()
        else:
            lime.empty()
        # About 0.4536 kg/hr of boiler chemicals are needed per 234484 kg/hr steam produced
        chems.imol['Ash'] = boiler_chems = 1.9345e-06 * Design['Flow rate']
        ash_disposal.empty()
        ash_disposal.copy_flow(emissions, IDs=tuple(ash_IDs), remove=True)
        ash_disposal.imol['Ash'] += boiler_chems
        dry_ash = ash_disposal.F_mass
        moisture = min(emissions.imass['Water'], dry_ash * 0.3) # ~20% moisture
        ash_disposal.imass['Water'] = moisture
        emissions.imass['Water'] -= moisture
        Design['Ash disposal'] = dry_ash + moisture
        if SO2_produced:
            if ID_lime == 'Ca(OH)2': # Ca(OH)2
                lime.imol['Water'] = 4 * lime_mol # Its a slurry
            else: # CaO
                lime.imol['Water'] = 5 * lime_mol 
        


# -*- coding: utf-8 -*-
"""
"""
import biosteam as bst
from math import sqrt, pi
import numpy as np

__all__ = ('FluidizedCatalyticCracking',)

class FluidizedCatalyticCracking(bst.Unit):
    """
    Create a reactor for catalytic cracking.
    
    Parameters
    ----------
    ins :
        * [0] liquid feed (upressurized)
        * [1] make-up catalyst
        * [2] air (unpressurized)
        * [3] steam (superheated)
    outs :
        * [0] liquid product
        * [1] spent catalyst
        * [2] flue gas
    
    Other Parameters
    ----------------
    ...
    
    Examples
    --------
    ...
    
    References
    ----------
    ...
    
    """
    _N_ins = 4
    _N_outs = 3
    auxiliary_unit_names = (
        'pump', 'feed_preheater', 'air_compressor', 'condenser', 
        'riser', 'stripper', 'regenerator', 'reactor',
    )

    def _init(
            self,
            reaction,
            bulk_reactant=None,
            vessel_material=None,
            feed_pressure=None,
            feed_vapor_fraction=None,
            catalyst_to_feed_ratio=None,
            spent_catalyst_to_feed_ratio=None,
            product_loss=None,
            riser_product_residence_time=None,
            # riser_entrance_superficial_velocity=None,
            riser_length_to_diameter=None,
            reactor_vessel_dissengagement_height=None, 
            reactor_vessel_exit_velocity=None,
            stripper_steam_rate=None,
            stripper_catalyst_residence_time=None,
            stripper_length_to_diameter=None,
            # regenerator_exit_velocity=None,
            CO2_concentration=None,
            regenerator_catalyst_residence_time=None,
            regenerator_length_to_diameter=None,
            regenerator_pressure=None,
        ):
        if vessel_material is None:
            vessel_material = 'Stainless steel 316'
        if feed_pressure is None:
            feed_pressure = 2 * 101325 # Can operate at a low pressure, but must accomodate pressure drop
        if feed_vapor_fraction is None:
            feed_vapor_fraction = 0
        if catalyst_to_feed_ratio is None:
            catalyst_to_feed_ratio = 5 # by weight
        if spent_catalyst_to_feed_ratio is None:
            # TODO: Find value
            # Mitias recommends (rule of thumb) 
            spent_catalyst_to_feed_ratio = 1e-6 # by weight; Value is abitrary for now 
        if product_loss is None:
            product_loss = 0.5e-2
        if riser_product_residence_time is None:
            riser_product_residence_time = 2.1 / 3600  # hr; 1.8 to 2.4 seconds preferred with VGO 
        # if riser_entrance_superficial_velocity is None:
        #     riser_entrance_superficial_velocity = 21945.6 # m / h; minimally 15 to 20 ft / s
        if riser_length_to_diameter is None:
            riser_length_to_diameter = 20 # Minimally preferred
        if reactor_vessel_dissengagement_height is None:
            reactor_vessel_dissengagement_height = 4.572 # m; 15 ft minimally
        if reactor_vessel_exit_velocity is None:
            reactor_vessel_exit_velocity = 2194.56 # m / s; 
        if stripper_steam_rate is None:
            stripper_steam_rate = 0.25 # wt %; 0.2-0.3 wt % steam per catalyst
        if stripper_catalyst_residence_time is None:
            stripper_catalyst_residence_time = 75 / 3600 # hr; typically 60 to 90 seconds
        if stripper_length_to_diameter is None:
            stripper_length_to_diameter = 1
        if CO2_concentration is None: 
            CO2_concentration = 0.055
        # if regenerator_exit_velocity is None:
        #     regenerator_exit_velocity = 2743.2 # m / h; between 2 to 3 ft / s
        if regenerator_catalyst_residence_time is None:
            regenerator_catalyst_residence_time = 7.5 / 60 # h; 5 to 10 min for 1-stage regenerator
        if regenerator_length_to_diameter is None:
            regenerator_length_to_diameter = 0.5 # Typically 0.3 to 0.7 
        if regenerator_pressure is None:
            regenerator_pressure = 282685 # 40 psig
        self.reaction = reaction
        self.bulk_reactant = bulk_reactant
        self.vessel_material = vessel_material
        self.feed_pressure = feed_pressure
        self.catalyst_to_feed_ratio = catalyst_to_feed_ratio
        self.spent_catalyst_to_feed_ratio = spent_catalyst_to_feed_ratio
        self.product_loss = product_loss
        self.riser_product_residence_time = riser_product_residence_time
        # self.riser_entrance_superficial_velocity = riser_entrance_superficial_velocity
        self.riser_length_to_diameter = riser_length_to_diameter
        self.reactor_vessel_dissengagement_height = reactor_vessel_dissengagement_height
        self.reactor_vessel_exit_velocity = reactor_vessel_exit_velocity
        self.stripper_steam_rate = stripper_steam_rate
        self.stripper_catalyst_residence_time = stripper_catalyst_residence_time
        self.stripper_length_to_diameter = stripper_length_to_diameter
        self.CO2_concentration = CO2_concentration
        # self.regenerator_exit_velocity = regenerator_exit_velocity
        self.regenerator_catalyst_residence_time = regenerator_catalyst_residence_time
        self.regenerator_length_to_diameter = regenerator_length_to_diameter
        self.regenerator_pressure = regenerator_pressure
        self.feed_vapor_fraction = feed_vapor_fraction
        
    @property
    def feed(self): return self.ins[0]
    
    @property
    def makeup_catalyst(self): return self.ins[1]
    
    @property
    def air(self): return self.ins[2]
    
    @property
    def steam(self): return self.ins[3]
    
    @property
    def product(self): return self.outs[0]
    
    @property
    def spent_catalyst(self): return self.outs[1]
    
    @property
    def flue_gas(self): return self.outs[2]
    
    def _setup(self):
        pump = self.auxiliary(
            'pump', bst.Pump, ins=self.feed, P=self.feed_pressure,
        )
        self.auxiliary(
            'feed_preheater', bst.HXutility, ins=pump-0, V=self.feed_vapor_fraction, rigorous=True,
        )
        self.auxiliary(
            'air_compressor', bst.IsentropicCompressor, ins=self.air, P=self.regenerator_pressure,
        )

    def _estimate_reactor_pressure_drop(self):
        return 40530.0

    def _estimate_regenerator_pressure_drop(self):
        return 40530.0

    def _run(self):
        feed, fresh_catalyst, air, steam = self.ins
        product, discarded_catalyst, flue_gas = self.outs
        self.pump.run()
        self.feed_preheater.run()
        F_mass_feed = feed.F_mass 
        fresh_catalyst.imass['Catalyst'] = discarded_catalyst.imass['Catalyst'] = F_mass_feed * self.spent_catalyst_to_feed_ratio
        self.catalyst_recirculation = catalyst_recirculation = F_mass_feed * self.catalyst_to_feed_ratio # kg / h
        steam.copy_like(bst.settings.get_heating_agent('low_pressure_steam'))
        steam.imass['Water'] = catalyst_recirculation * self.stripper_steam_rate
        product.P = self.feed_pressure - self._estimate_reactor_pressure_drop()
        flue_gas.P = discarded_catalyst.P = self.regenerator_pressure - self._estimate_regenerator_pressure_drop()
        product.mol = feed.mol + steam.mol
        if self.bulk_reactant:
            if self.reaction.X != 1:
                raise RuntimeError('reaction extent must be 100% with bulk reactants')
            model_reactant, reactants = self.bulk_reactant
            F_reactant_mass = product.imass[reactants].sum()
            product.imol[reactants] = 0
            product.imass[model_reactant] = F_reactant_mass
        product.phase = 'g'
        flue_gas.phase = 'g'
        self.reaction(product)
        # Product loss will dictate temperature of recirculated catalyst
        product.split_to(flue_gas, product, self.product_loss, energy_balance=False)
        product_loss = flue_gas.mol.copy()
        combustion = self.chemicals.get_combustion_reactions()
        combustion.force_reaction(flue_gas)
        O2 = -flue_gas.imass['O2']
        N2 = 0.79 / 0.21 * O2
        air.imass['O2', 'N2'] = [O2, N2]
        flue_gas.mol += air.mol
        F_emissions = flue_gas.F_mass
        z_CO2 = flue_gas.imass['CO2'] / F_emissions
        z_CO2_target = self.CO2_concentration
        if z_CO2 > z_CO2_target:
            F_emissions_new = z_CO2 * F_emissions / z_CO2_target
            dF_emissions = F_emissions_new - F_emissions
            air.F_mass = F_mass_O2_new = air.F_mass + dF_emissions
            flue_gas.mol += air.mol * (dF_emissions / F_mass_O2_new)
        self.air_compressor.run()
        regenerated_catalyst = bst.Stream(
            None, 
            Catalyst=catalyst_recirculation,
            units='kg/hr'
        )
        spent_catalyst = bst.Stream(
            None, 
            Catalyst=catalyst_recirculation, 
            units='kg/hr'
        )
        spent_catalyst.mol += product_loss
        compressed_air = self.air_compressor-0
        heated_feed = self.feed_preheater-0
        regenerator_outlets = [flue_gas, regenerated_catalyst, discarded_catalyst]
        regenerator_inlets = [compressed_air, spent_catalyst, fresh_catalyst]
        reactor_outlets = [product, spent_catalyst]
        reactor_inlets = [heated_feed, steam, regenerated_catalyst]
        feed_temperature = self.feed_preheater.outs[0].T
        for i in regenerator_outlets + reactor_outlets: i.T = feed_temperature # Initial guess
        dTs = np.ones(2)
        while np.abs(dTs).sum() > 0.1:
            # Energy balance correction equations
            # dT_regenerator * (C_flue_gas + C_catalyst_regen + C_discarded_catalyst) - dT_reactor * C_catalyst_spent = DeltaH_regen
            # dT_reactor * (C_catalyst_spent + C_product) - dT_regenerator * C_catalyst_regen = DeltaH_reactor
            A = np.array(
                [[sum([i.C for i in regenerator_outlets]), -spent_catalyst.C],
                 [-regenerated_catalyst.C, sum([i.C for i in reactor_outlets])]]
            )
            b = np.array(
                [sum([i.Hnet for i in regenerator_inlets]) - sum([i.Hnet for i in regenerator_outlets]),
                 sum([i.Hnet for i in reactor_inlets]) - sum([i.Hnet for i in reactor_outlets])]
            )
            dT_regenerator, dT_reactor = dTs = np.linalg.solve(A, b)
            for i in regenerator_outlets: i.T += dT_regenerator
            for i in reactor_outlets: i.T += dT_reactor

    def _design(self):
        # volume = L * pi * D**2 / 4
        # volume = D * L2D * pi * D**2 / 4
        # volume = D**3 * L2D * pi / 4
        # D = (volume * 4 / (pi * L2D)) ** (1 / 3)
        
        # Riser
        self.riser_volume = riser_volume = self.riser_product_residence_time * self.product.F_vol # m3
        L2D = self.riser_length_to_diameter
        self.riser_diameter = riser_diameter = (riser_volume * 4 / (pi * L2D)) ** (1 / 3)
        self.riser_length = riser_length = riser_diameter * L2D
        self.riser = bst.AuxiliaryPressureVessel(
            self.feed_pressure, riser_diameter, riser_length,
            pressure_units='Pa', length_units='m', 
            orientation='Vertical', material=self.vessel_material, 
        )
        
        # Reactor
        self.reactor_length = reactor_length = self.reactor_vessel_dissengagement_height
        self.reactor_diameter = reactor_diameter = sqrt(
            4 * self.product.F_vol / self.reactor_vessel_exit_velocity / pi
        )
        self.reactor = bst.AuxiliaryPressureVessel(
            self.feed_pressure, reactor_diameter, reactor_length,
            pressure_units='Pa', length_units='m', 
            orientation='Vertical', material=self.vessel_material, 
        )
        
        # Stripper
        catalyst_recirculation_volume = self.catalyst_recirculation / self.chemicals.Catalyst.rho(self.product.T, self.product.P)
        self.stripper_volume = stripper_volume = (
            self.stripper_catalyst_residence_time
            * catalyst_recirculation_volume
        )
        L2D = self.stripper_length_to_diameter
        self.stripper_diameter = stripper_diameter = (stripper_volume * 4 / (pi * L2D)) ** (1 / 3)
        self.stripper_length = stripper_length = stripper_diameter * L2D
        self.stripper = bst.AuxiliaryPressureVessel(
            self.feed_pressure, stripper_diameter, stripper_length,
            pressure_units='Pa', length_units='m', 
            orientation='Vertical', material=self.vessel_material, 
        )
        
        # Regenerator
        self.regenerator_volume = regenerator_volume = (
            self.regenerator_catalyst_residence_time
            * catalyst_recirculation_volume
        )
        L2D = self.regenerator_length_to_diameter
        self.regenerator_diameter = regenerator_diameter = (regenerator_volume * 4 / (pi * L2D)) ** (1 / 3)
        self.regenerator_length = regenerator_length = regenerator_diameter * L2D
        self.regenerator = bst.AuxiliaryPressureVessel(
            self.regenerator_pressure, regenerator_diameter, regenerator_length,
            pressure_units='Pa', length_units='m', 
            orientation='Vertical', material=self.vessel_material, 
        )
        
        for i in self.auxiliary_units: i._design()
    
    def _cost(self):
        for i in self.auxiliary_units: i._cost()


FCC = FluidizedCatalyticCracking


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from collections import namedtuple

__all__ = ('static',
           'static_flow_and_phase')

# %% Linking options


def _run_static(self):
    self._outs[0].copy_like(self._ins[0])

def static(cls=None):
    if cls is None: return lambda cls: static(cls)
    cls._link_streams = True
    cls._run = _run_static
    return cls

def _run_static_flow(self):
    outlet = self._outs[0]
    inlet = self._ins[0]
    outlet.phases = inlet.phases
    outlet.copy_flow(inlet)

def static_flow_and_phase(cls):
    cls._N_ins = cls._N_outs = 1
    if '_run' not in cls.__dict__: cls._run = cls._run_static_flow
    return cls

del namedtuple